import hashlib
import math
import time
from datetime import date, datetime, timezone
from python_graphql_client import GraphqlClient

from enum import Enum

from . import gql_mutations as gm
from . import gql_queries as gq

class NormalStatus(Enum):
    ENABLE = "ENABLE"
    DISABLE = "DISABLE"
    UNKNOWN__ = "UNKNOWN__"

class WatchOnlineStatus(Enum):
    UNKNOWN = "UNKNOWN"
    ONLINE = "ONLINE"
    OFFLINE = "OFFLINE"
    UNKNOWN__ = "UNKNOWN__"

class GQLHandler:
    def __init__(self, countryPhoneNumber: str, phoneNumber: str, password: str, userLang: str, timeZone: str):
        # init vars
        self.sessionId = None
        self.accessToken = None
        self.accessTokenExpire = 0
        self.userLocale = userLang
        self.timeZone = timeZone
        self.countryPhoneNumber = countryPhoneNumber
        self.phoneNumber = phoneNumber
        self.passwordMD5 = hashlib.md5(password.encode()).hexdigest()
        self.API_KEY = "270072d0fb4811ebacd96f6726fbdbb1"
        self.API_SECRET = "2d0288d0fb4811ebabfbd57e57c6ae64"
        self.issueDate = 0
        self.expireDate = 0
        self.userId = None
        self.ENDPOINT = "https://api.myxplora.com/api"
        self.variables = {
            "countryPhoneNumber": self.countryPhoneNumber,
            "phoneNumber": self.phoneNumber,
            "password": self.passwordMD5,
            "userLang": self.userLocale,
            "timeZone": self.timeZone
        }
        self.issueToken = None

    def getRequestHeaders(self, acceptedContentType: str):
        if acceptedContentType == "" or acceptedContentType == None:
            raise Exception("acceptedContentType MUST NOT be empty!")
        if self.API_KEY == None:
            raise Exception("Xplorao2o API_KEY MUST NOT be empty!")
        if self.API_SECRET == None:
            raise Exception("Xplorao2o API_SECRET MUST NOT be empty!")
        requestHeaders = {}
        if self.accessToken == None:
            # OPEN authorization
            authorizationHeader = f"Open {self.API_KEY}:{self.API_SECRET}"
        else:
            # BEARER authorization
            authorizationHeader = f"Bearer {self.accessToken}:{self.API_SECRET}"
            rfc1123DateString = datetime.now(timezone.utc).strftime("%a, %d %b %Y %H:%M:%S") + " GMT"
            requestHeaders["H-Date"] = rfc1123DateString
            requestHeaders["H-Authorization"] = authorizationHeader
        requestHeaders["H-BackDoor-Authorization"] = authorizationHeader
        requestHeaders["Accept"] = acceptedContentType
        requestHeaders["Content-Type"] = acceptedContentType
        requestHeaders["H-Tid"] = str(math.floor(time.time()))
        return requestHeaders

    async def runGqlQuery_a(self, query: str, variables):
        if query == None:
            raise Exception("GraphQL guery string MUST NOT be empty!")
        # Add Xplora® API headers
        requestHeaders = self.getRequestHeaders("application/json; charset=UTF-8")
        # create GQLClient
        gqlClient = GraphqlClient(endpoint=self.ENDPOINT, headers=requestHeaders)
        # execute QUERY|MUTATION
        data = await gqlClient.execute_async(query=query, variables=variables)
        return data

    async def runAuthorizedGqlQuery_a(self, query: str, variables):
        if self.accessToken == None:
            raise Exception("You have to login to the Xplora® API first.")
        # Run GraphQL query and return
        return (await self.runGqlQuery_a(query, variables))

    async def login_a(self):
        data = (await self.runGqlQuery_a(gm.MUTATION["tokenM"], self.variables))['data']
        if data['issueToken'] == None:
            # Login failed.
            raise LoginError("Login to Xplora® API failed. Check your input!")
        self.issueToken = data['issueToken']

        #  Login succeeded
        self.sessionId = self.issueToken['id'];
        self.userId = self.issueToken['user']['id'];
        self.accessToken = self.issueToken['token'];
        self.issueDate = self.issueToken['issueDate'];
        self.expireDate = self.issueToken['expireDate'];
    
        if self.issueToken['app'] != None:
            #  Update API_KEY and API_SECRET?
            if self.issueToken['app']['apiKey']:
                self.API_KEY = self.issueToken['app']['apiKey']
            if self.issueToken['app']['apiSecret']:
                self.API_SECRET = self.issueToken['app']['apiSecret']
        return self.issueToken

    async def isAdmin_a(self, ownId, query, variables, key):
        contacts = await self.getContacts_a(ownId)
        for contact in contacts['contacts']['contacts']:
            try:
                id = contact['contactUser']['id']
            except KeyError and TypeError:
                id = None
            if self.userId == id:
                if contact['guardianType'] == 'FIRST':
                    return (await self.runAuthorizedGqlQuery_a(query, variables))['data'][key]
        raise Exception("no Admin!")

########## SECTION QUERY start ##########

    async def getMyInfo_a(self): # Profil from login Account
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['readMyInfoQ'], {}))['data']

    async def getContacts_a(self, ownId): # Contacts from ownUser
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['contactsQ'], { "uid": ownId }))['data']

    async def getWatchCount_a(self):
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['watchCountQ'], {}))['data']

    async def getWatches_a(self, ownId):
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['watchesQ'], { "uid": ownId }))['data']

    async def getWatchLastLocation_a(self, ownId):
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['watchLastLocateQ'], { "uid": ownId }))['data']

    async def trackWatch_a(self, ownId): # tracking time - seconds
        res = (await self.runAuthorizedGqlQuery_a(gq.QUERY['trackWatchQ'], { "uid": ownId }))['data']
        if res['trackWatch'] != None:
            return res
        return { 'trackWatch': -1 }

    async def askWatchLocate_a(self, ownId):
        res = (await self.runAuthorizedGqlQuery_a(gq.QUERY['askWatchLocateQ'], { "uid": ownId }))['data']
        if res['askWatchLocate'] != None:
            return res
        return { 'askWatchLocate': False }

    async def getAlarms_a(self, ownId):
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['alarmsQ'], { "uid": ownId }))['data']

    async def getWifi_a(self, id): # without function?
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['getWifiQ'], { "uid": id }))['data']

    async def avatars_a(self, id): # without function?
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['avatarsQ'], { 'id': id }))['data']

    async def unReadChatMsgCount_a(self, ownId):
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['unReadChatMsgCountQ'], { 'uid': ownId }))['data']

    async def chats_a(self, ownId): # ownUser id
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['chatsQ'], { 'uid': ownId }))['data']

    async def notice_a(self): # without function?
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['noticeQ'], {}))['data']

    async def staticCard_a(self):
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['staticCardQ'], {}))['data']

    async def campaignUserProfiles_a(self):
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['CampaignUserProfilesQ'], {}))['data']

    async def subscribedCampaign_a(self, id, needDetail=False):
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['subscribedCampaignQ'], { 'uid': id, 'needDetail': needDetail }))['data']

    async def getReviewStatus_a(self, id):
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['GetReviewStatusQ'], { 'uid': id }))['data']

    async def countries_a(self): # Country Support
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['CountriesQ'], {}))['data']

    async def safeZones_a(self, ownId):
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['safeZonesQ'], { 'uid': ownId }))['data']

    async def safeZoneGroups_a(self):
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['safeZoneGroupsQ'], {}))['data']

    async def silentTimes_a(self, ownId):
        return (await self.runAuthorizedGqlQuery_a(gq.QUERY['silentTimesQ'], { 'uid': ownId }))['data']

########## SECTION QUERY end ##########
    
########## SECTION MUTATION start ##########

    async def sendText_a(self, ownId, text): # ownUser id
        if (await self.runAuthorizedGqlQuery_a(gm.MUTATION['sendTextM'], { "uid": ownId, "text": text }))['data']['sendChatText'] != None:
            return True
        return False

    async def addStep_a(self, stepCount):
        return (await self.runAuthorizedGqlQuery_a(gm.MUTATION['addStepM'], { "stepCount": stepCount }))['data']

    async def shutdown_a(self, ownId): # ownUser id
        return await self.isAdmin_a(ownId, gm.MUTATION['shutdownM'], { "uid": ownId }, 'reboot')

    async def reboot_a(self, ownId): # ownUser id
        return await self.isAdmin_a(ownId, gm.MUTATION['rebootM'], { "uid": ownId }, 'reboot')

    async def modifyAlert_a(self, id, YesOrNo): # function?
        return await self.runAuthorizedGqlQuery_a(gm.MUTATION['modifyAlertM'], { "uid": id, "remind": YesOrNo })

    async def setEnableSlientTime_a(self, silentId, status: NormalStatus=NormalStatus.ENABLE.value):
        return (await self.runAuthorizedGqlQuery_a(gm.MUTATION['setEnableSlientTimeM'], { 'silentId': silentId, 'status': status }))['data']

    async def setEnableAlarmTime_a(self, alarmId, status: NormalStatus=NormalStatus.ENABLE.value):
        return (await self.runAuthorizedGqlQuery_a(gm.MUTATION['ModifyAlarmM'], { 'alarmId': alarmId, 'status': status }))['data']

    async def setReadChatMsg(self, ownId, msgId, id):
        return (await self.runAuthorizedGqlQuery_a(gm.MUTATION['setReadChatMsg'], { "uid": ownId, 'msgId': msgId, 'id': id }))['data']

########## SECTION MUTATION end ##########

class Error(Exception):
    pass

class LoginError(Error):
    def __init__(self, message, res=1):
        self.message = message
        self.res = res
        super().__init__(self.message, self.res)

    def __str__(self):
        return f'{self.message} - {self.res}'


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "pyxplora_api",
    version = "1.0.63",
    author = "Ludy87",
    author_email = "android@astra-g.org",
    description = "Python Xplora® Api",
    license = "MIT",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/Ludy87/pyxplora_api",
    packages = setuptools.find_packages(),
    classifiers = [
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Topic :: Software Development",
        "Topic :: Utilities",
        "Topic :: Home Automation",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords = "api xplora watch",
    install_requires = [ "python-graphql-client==0.4.3" ],
)


from time import sleep
from .const import VERSION
from .gql_handler_async import *
from datetime import datetime
from asyncio import sleep

class PyXploraApi:
    def __init__(self, countrycode: str, phoneNumber: str, password: str, userLang: str, timeZone: str, watchNo: int=0) -> None:
        self._countrycode = countrycode
        self._phoneNumber = phoneNumber
        self._password = password
        self._userLang = userLang
        self._timeZone = timeZone
        self.watch_no = watchNo

        self.tokenExpiresAfter = 240
        self.maxRetries = 3
        self.retryDelay = 2

        self.dtIssueToken = int(time.time()) - (self.tokenExpiresAfter * 1000)

        self.contacts = []
        self.alarms = []
        self.chats = []
        self.safe_zones = []
        self.school_silent_mode = []
        self.watch_location = []
        self.watch_last_location = []
        self.watch_battery = None
        self.watch_charging = None

        self.__gqlHandler = None
        self.__issueToken = None
        self.watch = None
        self.user = None

    async def __login(self, forceLogin=False) -> dict:
        if not self.__isConnected() or self.__hasTokenExpired() or forceLogin:
            try:
                self.__logoff()
                self.__gqlHandler = GQLHandler(self._countrycode, self._phoneNumber, self._password, self._userLang, self._timeZone)
                if (self.__gqlHandler):
                    retryCounter = 0
                    while (not self.__isConnected() and (retryCounter < self.maxRetries + 2)):
                        retryCounter += 1

                        # Try to login
                        try:
                            self.__issueToken = await self.__gqlHandler.login_a();
                        except Exception:
                            pass

                        # Wait for next try
                        if (not self.__issueToken):
                            await sleep(self.retryDelay)
                    if (self.__issueToken):
                        self.dtIssueToken = int(time.time())
                else:
                    raise Exception("Unknown error creating a new GraphQL handler instance.")
            except Exception as error:
                # Login failed.
                self.__gqlHandler = None
                self.__issueToken = None
        return self.__issueToken

    def __isConnected(self) -> bool:
        return (self.__gqlHandler and self.__issueToken)

    def __logoff(self) -> None:
        self.__gqlHandler = None
        self.__issueToken = None

    def __hasTokenExpired(self) -> bool:
        return ((int(time.time()) - self.dtIssueToken) > (self.tokenExpiresAfter * 1000))

    async def init_async(self, forceLogin=False) -> None:
        token = await self.__login(forceLogin)
        if token:
            if ('user' in token):
                self.watch = token['user']['children'][self.watch_no]['ward']
                self.user = token['user']
                return
        raise Exception("Fail")

    def version(self) -> str:
        return VERSION

##### Contact Info #####
    async def getContacts_async(self) -> list:
        retryCounter = 0
        dataOk = False
        contacts_raw = None
        while (not dataOk and (retryCounter < self.maxRetries + 2)):
            retryCounter +=1
            await self.init_async()
            try:
                contacts_raw = await self.__gqlHandler.getContacts_a(await self.getWatchUserID_async())
                if 'contacts' in contacts_raw:
                    if 'contacts' in contacts_raw['contacts']:
                        for contact in contacts_raw['contacts']['contacts']:
                            try:
                                xcoin = contact['contactUser']['xcoin']
                                id = contact['contactUser']['id']
                            except TypeError:
                                xcoin = -1 # None - XCoins
                                id = None
                            self.contacts.append({
                                'id': id,
                                'guardianType': contact['guardianType'],
                                'create': datetime.fromtimestamp(contact['create']).strftime('%Y-%m-%d %H:%M:%S'),
                                'update': datetime.fromtimestamp(contact['update']).strftime('%Y-%m-%d %H:%M:%S'),
                                'name': contact['name'],
                                'phoneNumber': f"+{contact['countryPhoneNumber']}{contact['phoneNumber']}",
                                'xcoin': xcoin,
                            })
            except Exception as error:
                print(error)
            dataOk = self.contacts
            if (not dataOk):
                self.__logoff()
                await sleep(self.retryDelay)
        if (dataOk):
            return self.contacts
        else:
            raise Exception('Xplora API call finally failed with response: ')

##### User Info #####
    async def getUserID_async(self) -> str:
        return self.user['id']
    async def getUserName_async(self) -> str:
        return self.user['name']
    async def getUserIcon_async(self) -> str:
        return self.user['extra']['profileIcon']
    async def getUserXcoin_async(self) -> int:
        return self.user['xcoin']
    async def getUserCurrentStep_async(self) -> int:
        return self.user['currentStep']
    async def getUserTotalStep_async(self) -> int:
        return self.user['totalStep']
    async def getUserCreate_async(self) -> str:
        return datetime.fromtimestamp(self.user['create']).strftime('%Y-%m-%d %H:%M:%S')
    async def getUserUpdate_async(self) -> str:
        return datetime.fromtimestamp(self.user['update']).strftime('%Y-%m-%d %H:%M:%S')

##### Watch Info #####
    async def getWatchUserID_async(self) -> str:
        return self.watch['id']
    async def getWatchUserName_async(self) -> str:
        return self.watch['name']
    async def getWatchUserIcon_async(self) -> str:
        return f"https://api.myxplora.com/file?id={self.watch['file']['id']}"
    async def getWatchXcoin_async(self) -> int:
        return self.watch['xcoin']
    async def getWatchCurrentStep_async(self) -> int:
        return self.watch['currentStep']
    async def getWatchTotalStep_async(self) -> int:
        return self.watch['totalStep']

    async def getWatchAlarm_async(self) -> list:
        retryCounter = 0
        dataOk = False
        alarms_raw = None
        while (not dataOk and (retryCounter < self.maxRetries + 2)):
            retryCounter +=1
            await self.init_async()
            try:
                alarms_raw = await self.__gqlHandler.getAlarms_a(await self.getWatchUserID_async())
                if 'alarms' in alarms_raw:
                    if not alarms_raw['alarms']:
                        dataOk = True
                        return self.alarms
                    for alarm in alarms_raw['alarms']:
                        self.alarms.append({
                            'id': alarm['id'],
                            'name': alarm['name'],
                            'start': self.__helperTime(alarm['occurMin']),
                            'weekRepeat': alarm['weekRepeat'],
                            'status': alarm['status'],
                        })
            except Exception as error:
                print(error)
            dataOk = self.alarms
            if (not dataOk):
                self.__logoff()
                await sleep(self.retryDelay)
        if (dataOk):
            return self.alarms
        else:
            raise Exception('Xplora API call finally failed with response: ')
        return self.alarms

    async def loadWatchLocation_async(self, withAsk=True) -> list:
        retryCounter = 0
        dataOk = False
        location_raw = None
        while (not dataOk and (retryCounter < self.maxRetries + 2)):
            retryCounter +=1
            await self.init_async()
            try:
                if withAsk:
                    await self.askWatchLocate_async()
                await sleep(self.retryDelay)
                location_raw = await self.__gqlHandler.getWatchLastLocation_a(await self.getWatchUserID_async())
                if 'watchLastLocate' in location_raw:
                    if location_raw['watchLastLocate'] != None:
                        self.watch_location.append({
                            'tm': datetime.fromtimestamp(location_raw['watchLastLocate']['tm']).strftime('%Y-%m-%d %H:%M:%S'),
                            'lat': location_raw['watchLastLocate']['lat'],
                            'lng': location_raw['watchLastLocate']['lng'],
                            'rad': location_raw['watchLastLocate']['rad'],
                            'poi': location_raw['watchLastLocate']['poi'],
                            'city': location_raw['watchLastLocate']['city'],
                            'province': location_raw['watchLastLocate']['province'],
                            'country': location_raw['watchLastLocate']['country'],
                            'locateType': location_raw['watchLastLocate']['locateType'],
                            'isInSafeZone': location_raw['watchLastLocate']['isInSafeZone'],
                            'safeZoneLabel': location_raw['watchLastLocate']['safeZoneLabel'],
                        })
                        self.watch_battery = location_raw['watchLastLocate']['battery']
                        self.watch_charging = location_raw['watchLastLocate']['isCharging']
                        self.watch_last_location = location_raw['watchLastLocate']
            except Exception as error:
                print(error)
            dataOk = self.watch_location
            if (not dataOk):
                self.__logoff()
                await sleep(self.retryDelay)
        if (dataOk):
            return self.watch_location
        else:
            raise Exception('Xplora API call finally failed with response: ')

    async def getWatchBattery_async(self) -> int:
        await self.loadWatchLocation_async()
        return self.watch_battery
    async def getWatchIsCharging_async(self) -> bool:
        await self.loadWatchLocation_async()
        if self.watch_charging:
            return True
        return False
    async def getWatchOnlineStatus_async(self) -> WatchOnlineStatus:
        retryCounter = 0
        dataOk = False
        asktrack_raw = None
        while (not dataOk and (retryCounter < self.maxRetries + 2)):
            retryCounter +=1
            await self.init_async()
            try:
                await self.askWatchLocate_async()
                await sleep(self.retryDelay)
                ask_raw = await self.askWatchLocate_async()
                track_raw = await self.trackWatchInterval_async()
                if ask_raw or (track_raw != -1):
                    asktrack_raw = WatchOnlineStatus.ONLINE.value
                else:
                    asktrack_raw = WatchOnlineStatus.OFFLINE.value
            except Exception as error:
                print(error)
            dataOk = asktrack_raw
            if (not dataOk):
                self.__logoff()
                await sleep(self.retryDelay)
        if (dataOk):
            return asktrack_raw
        else:
            raise Exception('Xplora API call finally failed with response: ')
    async def __setReadChatMsg_a(self, msgId, id):
        return (await self.__gqlHandler.setReadChatMsg(await self.getWatchUserID_async(), msgId, id))['setReadChatMsg']
    async def getWatchUnReadChatMsgCount_async(self) -> int: # bug?
        return (await self.__gqlHandler.unReadChatMsgCount_a(await self.getWatchUserID_async()))['unReadChatMsgCount']
    async def getWatchChats_async(self) -> list: # bug?
        retryCounter = 0
        dataOk = False
        chats_raw = None
        while (not dataOk and (retryCounter < self.maxRetries + 2)):
            retryCounter +=1
            await self.init_async()
            try:
                await self.askWatchLocate_async()
                await sleep(self.retryDelay)
                chats_raw = await self.__gqlHandler.chats_a(await self.getWatchUserID_async())
                if 'chats' in chats_raw:
                    if 'list' in chats_raw['chats']:
                        for chat in chats_raw['chats']['list']:
                            self.chats.append({
                                'msgId': chat['msgId'],
                                'type': chat['type'],
                                #chat['sender'],
                                'sender_id': chat['sender']['id'],
                                'sender_name': chat['sender']['name'],
                                #chat['receiver'],
                                'receiver_id': chat['receiver']['id'],
                                'receiver_name': chat['receiver']['name'],
                                #chat['data'],
                                'data_text': chat['data']['text'],
                                'data_sender_name': chat['data']['sender_name'],
                            })
            except Exception as error:
                print(error)
            dataOk = self.chats
            if (not dataOk):
                self.__logoff()
                await sleep(self.retryDelay)
        if (dataOk):
            return self.chats
        else:
            raise Exception('Xplora API call finally failed with response: ')

##### Watch Location Info #####
    async def getWatchLastLocation_async(self, withAsk: bool = False) -> dict:
        await self.loadWatchLocation_async(withAsk)
        return self.watch_last_location
    async def getWatchLocate_async(self) -> dict:
        await self.loadWatchLocation_async()
        return self.watch_location
    async def getWatchLocateType_async(self) -> str:
        await self.getWatchLocate_async()
        return self.watch_last_location['locateType']
    async def getWatchIsInSafeZone_async(self) -> bool:
        await self.getWatchLocate_async()
        return self.watch_last_location['isInSafeZone']
    async def getWatchSafeZoneLabel_async(self) -> str:
        await self.getWatchLocate_async()
        return self.watch_last_location['safeZoneLabel']
    async def getSafeZones_async(self) -> list:
        retryCounter = 0
        dataOk = False
        safeZones_raw = None
        while (not dataOk and (retryCounter < self.maxRetries + 2)):
            retryCounter +=1
            await self.init_async()
            try:
                await self.askWatchLocate_async()
                await sleep(self.retryDelay)
                safeZones_raw = await self.__gqlHandler.safeZones_a(await self.getWatchUserID_async())
                if 'safeZones' in safeZones_raw:
                    for safeZone in safeZones_raw['safeZones']:
                        self.safe_zones.append({
                            #safeZone,
                            'vendorId': safeZone['vendorId'],
                            'groupName': safeZone['groupName'],
                            'name': safeZone['name'],
                            'lat': safeZone['lat'],
                            'lng': safeZone['lng'],
                            'rad': safeZone['rad'],
                            'address': safeZone['address'],
                        })
            except Exception as error:
                print(error)
            dataOk = self.safe_zones
            if (not dataOk):
                self.__logoff()
                await sleep(self.retryDelay)
        if (dataOk):
            return self.safe_zones
        else:
            raise Exception('Xplora API call finally failed with response: ')
    async def trackWatchInterval_async(self) -> int:
        return (await self.__gqlHandler.trackWatch_a(await self.getWatchUserID_async()))['trackWatch']
    async def askWatchLocate_async(self) -> bool:
        return (await self.__gqlHandler.askWatchLocate_a(await self.getWatchUserID_async()))['askWatchLocate']

##### Feature #####
    async def schoolSilentMode_async(self) -> list:
        retryCounter = 0
        dataOk = False
        sientTimes_raw = None
        while (not dataOk and (retryCounter < self.maxRetries + 2)):
            retryCounter +=1
            await self.init_async()
            try:
                await self.askWatchLocate_async()
                await sleep(self.retryDelay)
                sientTimes_raw = await self.__gqlHandler.silentTimes_a(await self.getWatchUserID_async())
                if 'silentTimes' in sientTimes_raw:
                    for sientTime in sientTimes_raw['silentTimes']:
                        self.school_silent_mode.append({
                            'id': sientTime['id'],
                            'start': self.__helperTime(sientTime['start']),
                            'end': self.__helperTime(sientTime['end']),
                            'weekRepeat': sientTime['weekRepeat'],
                            'status': sientTime['status'],
                        })
            except Exception as error:
                print(error)
            dataOk = self.school_silent_mode
            if (not dataOk):
                self.__logoff()
                await sleep(self.retryDelay)
        if (dataOk):
            return self.school_silent_mode
        else:
            raise Exception('Xplora API call finally failed with response: ')
    async def setEnableSilentTime_async(self, silentId) -> bool:
        retryCounter = 0
        dataOk = False
        _raw = None
        while (not dataOk and (retryCounter < self.maxRetries + 2)):
            retryCounter +=1
            await self.init_async()
            try:
                await self.askWatchLocate_async()
                await sleep(self.retryDelay)
                enable_raw = await self.__gqlHandler.setEnableSlientTime_a(silentId)
                if 'setEnableSilentTime' in enable_raw:
                    _raw = enable_raw['setEnableSilentTime']
            except Exception as error:
                print(error)
            dataOk = _raw
            if (not dataOk):
                self.__logoff()
                await sleep(self.retryDelay)
        if (dataOk):
            return bool(_raw)
        else:
            raise Exception('Xplora API call finally failed with response: ')
    async def setDisableSilentTime_async(self, silentId) -> bool:
        retryCounter = 0
        dataOk = False
        _raw = None
        while (not dataOk and (retryCounter < self.maxRetries + 2)):
            retryCounter +=1
            await self.init_async()
            try:
                await self.askWatchLocate_async()
                await sleep(self.retryDelay)
                disable_raw = await self.__gqlHandler.setEnableSlientTime_a(silentId, NormalStatus.DISABLE.value)
                if 'setEnableSilentTime' in disable_raw:
                    _raw = disable_raw['setEnableSilentTime']
            except Exception as error:
                print(error)
            dataOk = _raw
            if (not dataOk):
                self.__logoff()
                await sleep(self.retryDelay)
        if (dataOk):
            return bool(_raw)
        else:
            raise Exception('Xplora API call finally failed with response: ')
    async def setAllEnableSilentTime_async(self) -> list:
        res = []
        for silentTime in (await self.schoolSilentMode_async()):
            res.append(await self.setEnableSilentTime_async(silentTime['id']))
        return res
    async def setAllDisableSilentTime_async(self) -> list:
        res = []
        for silentTime in (await self.schoolSilentMode_async()):
            res.append(await self.setDisableSilentTime_async(silentTime['id']))
        return res

    async def setEnableAlarmTime_async(self, alarmId) -> bool:
        retryCounter = 0
        dataOk = False
        _raw = None
        while (not dataOk and (retryCounter < self.maxRetries + 2)):
            retryCounter +=1
            await self.init_async()
            try:
                await self.askWatchLocate_async()
                await sleep(self.retryDelay)
                enable_raw = await self.__gqlHandler.setEnableAlarmTime_a(alarmId)
                if 'modifyAlarm' in enable_raw:
                    _raw = enable_raw['modifyAlarm']
            except Exception as error:
                print(error)
            dataOk = _raw
            if (not dataOk):
                self.__logoff()
                await sleep(self.retryDelay)
        if (dataOk):
            return bool(_raw)
        else:
            raise Exception('Xplora API call finally failed with response: ')
    async def setDisableAlarmTime_async(self, alarmId) -> bool:
        retryCounter = 0
        dataOk = False
        _raw = None
        while (not dataOk and (retryCounter < self.maxRetries + 2)):
            retryCounter +=1
            await self.init_async()
            try:
                await self.askWatchLocate_async()
                await sleep(self.retryDelay)
                disable_raw = await self.__gqlHandler.setEnableAlarmTime_a(alarmId, NormalStatus.DISABLE.value)
                if 'modifyAlarm' in disable_raw:
                    _raw = disable_raw['modifyAlarm']
            except Exception as error:
                print(error)
            dataOk = _raw
            if (not dataOk):
                self.__logoff()
                await sleep(self.retryDelay)
        if (dataOk):
            return bool(_raw)
        else:
            raise Exception('Xplora API call finally failed with response: ')
    async def setAllEnableAlarmTime_async(self) -> list:
        res = []
        for alarmTime in (await self.getWatchAlarm_async()):
            res.append(await self.setEnableAlarmTime_async(alarmTime['id']))
        return res
    async def setAllDisableAlarmTime_async(self) -> list:
        res = []
        for alarmTime in (await self.getWatchAlarm_async()):
            res.append(await self.setDisableAlarmTime_async(alarmTime['id']))
        return res

    async def sendText(self, text) -> bool: # sender is login User
        return await self.__gqlHandler.sendText_a(await self.getWatchUserID_async(), text)
    async def isAdmin(self) -> bool:
        for contact in await self.getContacts_async():
            if (contact['id'] == await self.getUserID_async()):
                return True
        return False
    async def shutdown(self) -> bool:
        if self.isAdmin():
            return await self.__gqlHandler.shutdown_a(await self.getWatchUserID_async())
        raise Exception("no Admin")
    async def reboot(self) -> bool:
        if self.isAdmin():
            return await self.__gqlHandler.reboot_a(await self.getWatchUserID_async())
        raise Exception("no Admin")

##### - #####
    def __helperTime(self, time) -> str:
        h = str(int(time) /60).split('.')
        h2 = str(int(h[1]) *60).zfill(2)[:2]
        return h[0].zfill(2) + ":" + str(h2).zfill(2)
